import numpy as np
import pandas as pd
import statsmodels.formula.api as smf
import matplotlib.pyplot as plt
import seaborn as sb

def calibrate_band(model, sample, u_trim, smoothing):
    sample_size = len(sample)
    std = np.std(sample)
    band = 1.06*std*np.power(sample_size, smoothing)
    delta = 1 # this is only if on [0,1]
    u_band = band/delta
    i_band = int(u_band*sample_size)
    trim = int(u_trim*sample_size)

    if trim < i_band:
        print('Warning: Not enough trimming, look out for boundary effects.')

    return sample_size, band, i_band, trim
    
def calibrate_part(model, u_grid, frec):

    M = np.max(list(frec.keys()))
    m_min = np.min(list(frec.keys()))

    A_1 = 0*u_grid
    A_1_prime = 0*u_grid
    a = 0

    for m, pm in frec.items():
        A_1 += pm*np.power(u_grid, m-1)
        A_1_prime += pm*(m-1)*np.power(u_grid, m-2)
        a += m*pm/M

    A_1[0] = A_1[1] # avoid division by zero
    A_1_prime[0] = A_1_prime[1] # avoid division by zero

    A_2 = u_grid*A_1
    A_3 = (1-u_grid)*A_1

    A_4 = A_1/A_1_prime

    return M, A_1, A_2, A_3, A_4, a


import numpy as np
from scipy.stats import norm
import os
from multiprocess import Pool
#from pathos.multiprocessing import ProcessingPool as Pool

from .estimators import *

def make_ci(self, confidence, hyp):
    
    one = norm.ppf(confidence/100)
    two = norm.ppf((confidence+(100-confidence)/2)/100)
    
    self.ci_one = np.sqrt(self.intKsq)*one
    self.ci_two = np.sqrt(self.intKsq)*two
    
    if hyp == 'twosided':
        self.ci = self.ci_two

    self.core_ci = self.ci*self.hat_q/np.sqrt(self.sample_size*self.band)
        
    self.data['_q_ci'] = np.nan
    self.data.loc[self.active_index,'_q_ci'] = self.core_ci

    self.data['_v_ci'] = np.nan
    self.data.loc[self.active_index,'_v_ci'] = self.A_4*self.core_ci
    
    self.data['_bs_ci'] = np.nan
    self.data.loc[self.active_index,'_bs_ci'] = self.a*self.A_3*self.A_4*self.core_ci
    
    self.data['_rev_ci'] = np.nan
    self.data.loc[self.active_index,'_rev_ci'] = self.M*self.a*self.A_3*self.A_4*self.core_ci
    
def make_cb(self, confidence, draws, hyp):

    def simulate_q(i): 
        np.random.seed(i)
        mc = np.sort(np.random.uniform(0, 1, self.sample_size))
        return q_smooth(mc, self.kernel, *self.band_options, reflect = True, is_sorted = True)

    p = Pool(os.cpu_count())
    hat_qs = np.array(p.map(simulate_q, range(draws)))
    p.close()
    p.join()
    
    sup = np.max((hat_qs-1)[:,self.trim:-self.trim], axis = 1)
    supabs = np.max(np.abs(hat_qs-1)[:,self.trim:-self.trim], axis = 1)

    self.cb_one = np.percentile(sup, confidence)
    self.cb_two = np.percentile(supabs, confidence+(100-confidence)/2)
        
    if hyp == 'twosided':
        self.cb = self.cb_two
        
    self.core_cb = self.cb*self.hat_q
        
    self.data['_q_cb'] = np.nan
    self.data.loc[self.active_index,'_q_cb'] = self.core_cb

    self.data['_v_cb'] = np.nan
    self.data.loc[self.active_index,'_v_cb'] = self.A_4*self.core_cb
    
    self.data['_bs_cb'] = np.nan
    self.data.loc[self.active_index,'_bs_cb'] = self.a*self.A_3*self.A_4*self.core_cb
    
    self.data['_rev_cb'] = np.nan
    self.data.loc[self.active_index,'_rev_cb'] = self.M*self.a*self.A_3*self.A_4*self.core_cb
    
def make_cicb_for_ts(self, confidence, draws, hyp):
    
    def simulate_ts(i): 
        np.random.seed(i)
        hat_F = np.sort(np.random.uniform(0, 1, self.sample_size))
        return total_surplus_from_Q(self.hat_q*hat_F, *self.part_options)
    
    p = Pool(os.cpu_count())
    hat_TSs = np.array(p.map(simulate_ts, range(draws)))
    p.close()
    p.join()
    
    right = total_surplus_from_Q(self.hat_q*self.u_grid, *self.part_options)
    
    sup_ts = np.max((hat_TSs-right)[:,self.trim:-self.trim], axis = 1)
    abs_ts = np.abs(hat_TSs-right)
    supabs_ts = np.max(abs_ts[:,self.trim:-self.trim], axis = 1)
    
    self.ci_two_ts = np.percentile(abs_ts, confidence+(100-confidence)/2, axis = 0)
    
    self.cb_one_ts = np.percentile(sup_ts, confidence)
    self.cb_two_ts = np.percentile(supabs_ts, confidence+(100-confidence)/2)
    
    if hyp == 'twosided':
        self.cb_ts = self.cb_two_ts
        self.ci_ts = self.ci_two_ts

    self.data['_ts_ci'] = np.nan
    self.data.loc[self.active_index,'_ts_ci'] = self.ci_ts
        
    self.data['_ts_cb'] = np.nan
    self.data.loc[self.active_index,'_ts_cb'] = self.cb_ts
    



import numpy as np
import pandas as pd
import statsmodels.formula.api as smf
import matplotlib.pyplot as plt
import seaborn as sb
from scipy.stats import norm
import pkg_resources

from .kernels import *
from .estimators import *
from .plots import *
from .calibrate import *
from .inference import *

from multiprocess import Pool
import numba as nb
    
def cens_Q(Q, x, eps):
    return (Q(eps+x*(1-2*eps))-Q(eps))/(Q(1-eps)-Q(eps))

def cens_q(f, Q, x, eps):
    return (1-2*eps)/(f(Q(eps+x*(1-2*eps)))*(Q(1-eps)-Q(eps)))

class Simulator:
    '''Addition to the package for the "Nonparametric inference on counterfactuals in sealed first-price auctions" paper.'''
    
    def __init__(self, sample_size, smoothing_rate, trim_percent, 
                 frec, rvpdf, rvppf, eps, draws):
        
        self.u_grid = np.linspace(0, 1, sample_size)
        
        self.Q_fun = lambda x: cens_Q(rvppf, x, eps)
        self.q_fun = lambda x: cens_q(rvpdf, rvppf, x, eps)
        
        self.sample_size = sample_size
        self.frec = frec
        
        self.smoothing = -smoothing_rate
        self.u_trim = trim_percent/100
        
        self.draws = draws
        
    def calibrate(self):
        
        self.mc = self.Q_fun(np.sort(np.random.uniform(0, 1, size = self.sample_size)))
        
        self.band_options = calibrate_band(self, self.mc, self.u_trim, self.smoothing)
        self.sample_size, self.band, self.i_band, self.trim = self.band_options
        
        self.kernel, _ = make_kernel(self.u_grid, self.i_band, kernel = tri)
        
        self.part_options = calibrate_part(self, self.u_grid, self.frec)
        self.M, self.A_1, self.A_2, self.A_3, self.A_4, self.a = self.part_options
        
    def simulate(self):
        
        trim = self.trim
        draws = self.draws
        M = self.M
        A_2 = self.A_2
        A_3 = self.A_3
        A_4 = self.A_4
        a = self.a
        u_grid = self.u_grid
        
        true_Q = self.Q_fun(self.u_grid)
        true_q = self.q_fun(self.u_grid)
        
        # eraze boundary
        true_Q[-self.trim:] = 0
        true_q[-self.trim:] = 0
        
        def one_mc(i):
            np.random.seed(i)

            Q_uni = np.sort(np.random.uniform(0, 1, self.sample_size))
            Q_dgp = self.Q_fun(Q_uni)

            q_uni = q_smooth(Q_uni, self.kernel, *self.band_options, is_sorted = True, reflect = False)
            q_dgp = q_smooth(Q_dgp, self.kernel, *self.band_options, is_sorted = True, reflect = False)

            # eraze boundary
            Q_uni[-self.trim:] = 0
            q_uni[-self.trim:] = 0
            Q_dgp[-self.trim:] = 0
            q_dgp[-self.trim:] = 0

            return [Q_uni, q_uni, Q_dgp, q_dgp]
        
        p = Pool(os.cpu_count())
        all_mc = np.array(p.map(one_mc, range(self.draws)))
        p.close()
        p.join()
        
        all_Q_uni  = all_mc[:,0,:].copy()
        all_q_uni  = all_mc[:,1,:].copy()

        all_Q_dgp  = all_mc[:,2,:].copy()
        all_q_dgp  = all_mc[:,3,:].copy()

        del(all_mc)
        
        @nb.jit(nopython = True)
        def d_numba(arr):
            diff = arr - np.roll(arr, 1)
            diff[0] = diff[1]
            return diff*len(diff)

        @nb.jit(nopython = True)
        def int_lowbound_numba(arr):
            return np.flip(np.cumsum(np.flip(arr)))/len(arr)
        
        psi = d_numba(A_2)
        chi = psi - d_numba(A_4*psi)
        
        stats_dgp = np.zeros(shape = (draws, 5), dtype = np.float)
        
        @nb.jit(nopython = True, parallel = True)
        def simulate_all_dgp(stats_dgp):
            for i in nb.prange(draws):
                delta_Q = all_Q_dgp[i]-true_Q
                delta_q = all_q_dgp[i]-true_q

                delta_v = delta_Q + A_4*delta_q
                delta_ts = int_lowbound_numba(delta_v*d_numba(A_2))
                delta_bs = a*int_lowbound_numba(A_3*d_numba(delta_v))
                delta_rev = delta_ts - M*delta_bs

                # this is a better way
                delta_ts = A_4[-1]*psi[-1]*delta_Q[-1]-A_4*psi*delta_Q + int_lowbound_numba(chi*delta_Q)

                stats_dgp[i,0] = np.max(np.abs(delta_q)[trim:-trim])
                stats_dgp[i,1] = np.max(np.abs(delta_v)[trim:-trim])
                stats_dgp[i,2] = np.max(np.abs(delta_bs)[trim:-trim])
                stats_dgp[i,3] = np.max(np.abs(delta_rev)[trim:-trim])
                stats_dgp[i,4] = np.max(np.abs(delta_ts)[trim:-trim])
                
        simulate_all_dgp(stats_dgp)
        self.stats_dgp = stats_dgp
        
        stats_uni = np.zeros(shape = (draws, draws, 5), dtype = np.float)

        @nb.jit(nopython = True, parallel = True)
        def simulate_all_uni(stats_uni):
            for i in nb.prange(draws):
                for j in nb.prange(draws):
                    delta_Q = (all_Q_uni[j]-u_grid)*all_q_dgp[i]
                    delta_q = (all_q_uni[j]-1)*all_q_dgp[i]

                    delta_v = delta_Q + A_4*delta_q
                    delta_ts = int_lowbound_numba(delta_v*d_numba(A_2))
                    delta_bs = a*int_lowbound_numba(A_3*d_numba(delta_v))
                    delta_rev = delta_ts - M*delta_bs

                    # this is a better way
                    delta_ts = A_4[-1]*psi[-1]*delta_Q[-1]-A_4*psi*delta_Q + int_lowbound_numba(chi*delta_Q)

                    stats_uni[i,j,0] = np.max(np.abs(delta_q)[trim:-trim])
                    stats_uni[i,j,1] = np.max(np.abs(delta_v)[trim:-trim])
                    stats_uni[i,j,2] = np.max(np.abs(delta_bs)[trim:-trim])
                    stats_uni[i,j,3] = np.max(np.abs(delta_rev)[trim:-trim])
                    stats_uni[i,j,4] = np.max(np.abs(delta_ts)[trim:-trim])
                    
        simulate_all_uni(stats_uni)
        self.stats_uni = stats_uni
        
    def get_coverage(self, nominal_coverage, digits):
        crit_uni = np.percentile(self.stats_uni.reshape(self.draws, self.draws, 5), nominal_coverage, axis = 1)
        return {i:np.round((np.mean(np.sign(crit_uni[:,j]-self.stats_dgp[:,j]))+1)/2, digits) for i,j in zip(['q','v','bs','rev','ts'],range(5))}
        
        